
xVersion" 2011.11.5.1"

module Memory

  #document Memory
    //
    // Flash memory interface
    //
    // Uses a configurable portion of the processor's FLASH memory as a user
    // data storage device.
    //
    // The portion to be used should start on an erase page boundary, but to
    // be able to write protect the other areas of the flash independently
    // the portion should start on a panel / block boundary address.
    // This code is meant for the 24xxxx family of processors which has a
    // block size of 16 kiBytes.
    //
    // To modify a bit of memory the parts to be modified should be read into
    // RAM where they are modified. This has to be a page sized chunk of memory
    // (64 instructions, 192 bytes). Then the page in flash has to be erased
    // and then the RAM contents can be written to flash.
    //
  #endDoc

private

  uses
    hw_interface utils
  endUses

  // The memory addresses usd below are instruction addresses.

 2variable MemStorageStart          #// Start offset in flash for user data storage
                                    #// everything below it is program memory
                                    #// multiple of 64 instructions (192 bytes),
                                    #// preferrably on a panel boundary, but it
                                    #// needs be an even address at least. This
                                    #// is an instruction offset.

  variable MemStorageSize           #// Size of user storage area, upto 64 kib
                                    #// (minus one byte, 0 means do not use it)


 dvarInterface  MemStorageStart
  varInterface  MemStorageSize



 512 constant MemErasePageSize      #// Size of a memory erase page in instruction locations (1536 bytes)
  64 constant MemWritePageSize      #// Size of a memory write page in instruction locations ( 192 bytes)



  // Flash opcodes
                                     // fedc ba98 7654 3210
  $4042 constant  FLASH_PAGE_ERASE  #// 0100 0000 0100 0010  see [[FLASH_DISABLE]] for bit defs
  $4001 constant  FLASH_ROW_WRITE   #// 0100 0000 0000 0001  see [[FLASH_DISABLE]] for bit defs
  $0000 constant  FLASH_DISABLE     #// 0000 0000 0000 0000  also see [[FLASH_PAGE_ERASE]] and [[FLASH_ROW_WRITE]]
                                    #// |||| |||| |||| ||||   3210 NVMOP
                                    #// |||| |||| |||| ++++-- 1111 - bulk erase, only available for ICSP mode
                                    #// |||| |||| ||||        0011 - mem word program mode when ERASE = 0
                                    #// |||| |||| ||||        0010 - mem page erase        when ERASE = 1
                                    #// |||| |||| ||||        0001 - mem row program       when ERASE = 0
                                    #// |||| |||| ||||
                                    #// |||| |||| ||++------- unimplemented, read as 0
                                    #// |||| |||| ||
                                    #// |||| |||| ||          6 ERASE
                                    #// |||| |||| |+--------- 1 - preform erase op specified in NVMOP
                                    #// |||| |||| |           0 - perform write op specified in NVMOP
                                    #// |||| |||| |
                                    #// |||+-++++-+---------- unimplemented, read as 0
                                    #// |||
                                    #// |||                   d WRERR
                                    #// ||+------------------ 1 - improper write or erase occurred
                                    #// ||                    0 - program or erase completed OK
                                    #// ||
                                    #// ||                    e WREN
                                    #// |+------------------- 1 - enable  flash / erase ops
                                    #// |                     0 - disable flash / erase ops
                                    #// |
                                    #// |                     f WR
                                    #// +-------------------- 1 - initiate flash write / erase
                                    #// .                     0 - flash write / erase completed


  // //////////////////////////////////////////////////////////////////////////
  //
  // Low level memory access stuff


  code MemEraseBlock ( dRomOffs -- )

    push      TBLPAG                ; // Save TBLPAG register
    PopReg    w1                    ; // dRomOffs into w1:w0

    mov       w1, TBLPAG            ; // Init Pointer to page to be erased
    tblwtl    w0, [w0]              ; // Dummy write to select the row

    mov       #<@" FLASH_PAGE_ERASE">, w2  ; // Specify page erase operation
    mov       w2, NVMCON
    bset      w2, #WR               ; // Specify action to occur

    disi      #5                    ; // Block interrupt with priority < 7 for next 5 instructions

    mov       #0x55, w0             ; // Unlock sequence - interrupts need to be off
    mov       w0, NVMKEY
    mov       #0xaa, w0
    mov       w0, NVMKEY
    mov       w2, NVMCON            ; // Start the specified action

    nop                             ; // Required nops
    nop

1:

    btsc      NVMCON, #WR           ; // Wait for write end
    bra       1b

    mov       #<@" FLASH_DISABLE">, w2  ; // Disable flash ops
    mov       w2, NVMCON

    drop                            ; // drop last byte of dRomOffs
    pop       TBLPAG                ; // Restore TBLPAG register
    return                          ; // ----> done
  endCode
  #document MemEraseBlock
    //
    // Erases an erase page worth of flash memory, dRomOffs
    // can be anywhere within the block, all 512 instructions
    // (1536) bytes will be erased.
    //
    // A program memory page erase operation is set up by performing a dummy table
    // write (TBLWTL) operation to any address within the page.
    //
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  code MemWriteBlock  ( nRamAddrSrc dRomAddrDst uCount -- ) // Copy RAM data to ROM, uCount bytes

    push      TBLPAG                ; // Don't change TBLPAG

    PopReg    w3                    ; // Length             into w3
    PopReg    w1                    ; // High ROM address   into w1
    PopReg    w4                    ; // Low  ROM address   into w4
    mov       w0, w2                ; // Src address in RAM into w2
    mov       w4, w0                ; // Low  ROM address   into w0

1:
    dec       w3, w3                ; // num--
    bra       n, 2f                 ; // B/ done

    mov       w1, TBLPAG            ; // Point to write page

    tblwtl.b  [w2++], [w0++]        ; // dst++.lhb = lo byte

    dec       w3, w3                ; // num--
    bra       n, 2f                 ; // B/ done

    tblwtl.b  [w2++], [w0--]        ; // dst--.llb = hi byte

    dec       w3, w3                ; // num--
    bra       n, 2f                 ; // B/ done

    tblwth.b  [w2++], [w0]          ; // dst.hlb = upper byte

    add       w0, #2, w0            ; // dst += 2
    addc      w1, #0, w1

    bra       1b                    ; // B/ do next
2:
    mov       #<@" FLASH_ROW_WRITE">, w2  ; // Specify page write operation
    mov       w2, NVMCON
    bset      w2, #WR               ; // Specify action to occur

    disi      #5                    ; // Block interrupt with priority < 7 for next 5 instructions

    mov       #0x55, w0             ; // Unlock sequence - interrupts need to be off
    mov       w0, NVMKEY
    mov       #0xaa, w0
    mov       w0, NVMKEY
    mov       w2, NVMCON            ; // Start the specified action

    nop                             ; // Required nops
    nop

3:

    btsc      NVMCON, #WR           ; // Wait for write end
    bra       3b

    mov       #<@" FLASH_DISABLE">, w2    ; // Disable flash ops
    mov       w2, NVMCON

    drop                            ; // Loose last data stack item
    pop       TBLPAG                ; // Restore TBLPAG to entry value

    return                          ; // B/ ----> done

  endCode
  #document MemWriteBlock
    //
    // Writes nCount bytes of flash memory, dRomAddr need not point
    // to the start of an erase page, but must point to an even address
    // and the operation is not allowed to go over an erase page, eg
    // dRomAddr + nCount - 1 must be in the same MemWritePageSize block as
    // dRomAddr + 0. To be able to set bits from a zero to a 1 the page
    // should have been erased before writing (the only way to make 1s).
    // The maximum block size is MemWritePageSize * 3 bytes (uCount),
    // provided the start address is at the start of a write page.
    //
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  code MemReadBlock ( nRamAddrDst dRomAddrSrc uCount -- ) // Copy ROM data to RAM, uCount bytes

    push      TBLPAG                ; // Don't change TBLPAG

    PopReg    w3                    ; // Length             into w3
    PopReg    w1                    ; // High ROM address   into w1
    PopReg    w4                    ; // Low  ROM address   into w4
    mov       w0, w2                ; // Dst address in RAM into w2
    mov       w4, w0                ; // Low  ROM address   into w0

1:
    dec       w3, w3                ; // num--
    bra       n, 2f                 ; // B/ done

    mov       w1, TBLPAG            ; // TBLPAG = tblpage  ( src)

    tblrdl.b  [w0++], [w2++]        ; // dst++ = lo byte

    dec       w3, w3                ; // num--
    bra       n, 2f                 ; // B/ done

    tblrdl.b  [w0--], [w2++]        ; // dst++ = hi byte

    dec       w3, w3                ; // num--
    bra       n, 2f                 ; // B/ done

    tblrdh.b  [w0], [w2++]          ; // dst++ = upper byte

    add       w0, #2, w0            ; // src += 2
    addc      w1, #0, w1

    bra       1b                    ; // B/ do next
2:
    drop                            ; // Loose last data stack item
    pop       TBLPAG                ; // Restore TBLPAG to entry value

    return                          ; // B/ ----> done

  endCode
  #document MemReadBlock
    //
    // Reads nCount bytes of flash memory int RAM, dRomAddr need not point
    // to the start of an erase or write page, more than MemErasePageSize *
    // 3 bytes can be read (up to 65536).
    //
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  code MemVerifyBlock ( nRamAddr dRomOffs nCount -- f )

    push      TBLPAG                ; // Don't change TBLPAG

    PopReg    w3                    ; // Length             into w3
    PopReg    w1                    ; // High ROM address   into w1
    PopReg    w4                    ; // Low  ROM address   into w4
    mov       w0, w2                ; // Dst address in RAM into w2
    mov       w4, w0                ; // Low  ROM address   into w0

1:
    dec       w3, w3                ; // num--
    bra       n, 2f                 ; // B/ done

    mov       w1, TBLPAG            ; // TBLPAG = tblpage  ( src)

    tblrdl.b  [w0++], w4            ; // w4 = rom++
    cp.b      w4, [w2++]            ; // cp against dst++
    bra       nz, 3f                ; // B/ compare error

    dec       w3, w3                ; // num--
    bra       n, 2f                 ; // B/ done

    tblrdl.b  [w0--], w4            ; // w4 = rom++
    cp.b      w4, [w2++]            ; // cp against dst++
    bra       nz, 3f                ; // B/ compare error

    dec       w3, w3                ; // num--
    bra       n, 2f                 ; // B/ done

    tblrdh.b  [w0], w4              ; // w4 = rom
    cp.b      w4, [w2++]            ; // cp against dst++
    bra       nz, 3f                ; // B/ compare error

    add       w0, #2, w0            ; // src += 2
    addc      w1, #0, w1

    bra       1b                    ; // B/ do next

2:                                  ; // All compared OK
    mov       #True, w0             ; // Signal comparison to be OK
    bra       4f                    ; // B/ done

3:                                  ; // Compare error

    mov       #False, w0            ; // Signal compare error

4:                                  ; // Done
    pop       TBLPAG                ; // Restore TBLPAG to entry value

    return                          ; // B/ ----> done

  endCode
  #document MemVerifyBlock
    //
    // Reads nCount bytes of flash memory and compares them to RAM, dRomAddr
    // need not point to the start of an erase or write page.
    // When all FLASH an RAM bytes have the same value True is returned, False
    // otherwise. More than MemErasePageSize * 3 bytes can be verified (up to 65536).
    //
  #endDoc


  // //////////////////////////////////////////////////////////////////////////
  //
  // : MemUserEraseAllStorage ( -- )
  //
  //   #// Completely erase the user storage area
  //   #// This is [[MemStorageStart]] up to and including [[MemStorageEnd]]
  //
  //   [                                     // Calculate nr. of pages to erase.
  //     MemStorageEnd 1 +                   // Erase will start at the end of the
  //     MemStorageStart -                   // user memory area and we start one
  //     MemErasePageSize /                  // page too high so we need pre-
  //     dLiteral                            // decrement in the loop that we can
  //                                         // terminate on zero.
  //   ]
  //   begin
  //     ddup d0<>                           // As long as page number <> 0
  //   while
  //     KickDog                             // Keep the watchdog from kicking
  //                                         // typical self timed write takes 2ms
  //                                         // typical. So typically we need > 2s
  //                                         // here. The dog is set >= ?? ms.
  //
  //     1d-                                 // pre-decrement page number
  //     2dup                                // Dup page number
  //     MemErasePageSize                    // Multiply by page size
  //     u>d d*
  //     MemUserEraseBlock                   // Erase the block
  //   repeat
  //   2Drop                                 // Drop page number
  // ;
  //
  // //////////////////////////////////////////////////////////////////////////

  : MemReadByte ( dRomOffs -- n )

    #// Read a byte from memory

    RomPtr!
    Romc@
  ;

  // //////////////////////////////////////////////////////////////////////////

  : MemReadWord ( dRomOffs -- n )

    #// Read word from memory

    RomPtr!
    2 0 do Romc@++ loop
  ;

  // //////////////////////////////////////////////////////////////////////////

  : MemReadTripple ( dRomOffs -- d )

    #// Read three bytes from memory as a double

    RomPtr!
    3 0 do Romc@++ loop
  ;

  // //////////////////////////////////////////////////////////////////////////

  : MemReadQuad ( dRomOffs -- d )

    #// Read a double from memory

    RomPtr!
    4 0 do Romc@++ loop
  ;


  // //////////////////////////////////////////////////////////////////////////

  : MemMakePhysical ( dRomOffs -- dRomAddr )  #// Adds MemStorageStart to dRomAddr

    MemStorageStart@ d+
  ;


  // //////////////////////////////////////////////////////////////////////////

  : MemUserEraseBlock ( dRomOffs -- )

    #// Erase a user memory block

    MemMakePhysical
    MemEraseBlock
  ;


  // //////////////////////////////////////////////////////////////////////////

  : MemUserWriteBlock ( nRamAddr dRomOffs nCount -- )

    #// Write a block from RAM into user memory

    >r
    MemMakePhysical
    r>
    MemWriteBlock
  ;


  // //////////////////////////////////////////////////////////////////////////

  : MemUserReadBlock ( nRamAddr dRomOffs nCount  -- )

    #// Read a block from user memory to RAM

    >r
    MemMakePhysical
    r>
    MemReadBlock
  ;


  // //////////////////////////////////////////////////////////////////////////

  : MemUserVerifyBlock ( nRamAddr dRomOffs nCount -- f )

    #// Read a block from user memory and compare to RAM

    >r
    MemMakePhysical
    r>
    MemVerifyBlock
  ;


  // //////////////////////////////////////////////////////////////////////////

  : MemUserReadByte ( dRomOffs -- n )

    #// Read a byte from user memory

    MemMakePhysical
    MemReadByte
  ;

  // //////////////////////////////////////////////////////////////////////////

  : MemUserReadWord ( dRomOffs -- n )

    #// Read word from user memory

    MemMakePhysical
    MemReadWord
  ;

  // //////////////////////////////////////////////////////////////////////////

  : MemUserReadTripple ( dRomOffs -- d )

    #// Read three bytes from user memory as a double

    MemMakePhysical
    MemReadTripple
  ;

  // //////////////////////////////////////////////////////////////////////////

  : MemUserReadQuad ( dRomOffs -- d )

    #// Read a double from user memory

    MemMakePhysical
    MemReadQuad
  ;

  // //////////////////////////////////////////////////////////////////////////

  : InitMemoryModule ( dRomOffset nCount -- )

    #// Module initialization
    // use MemStroageStart = $01.0000 to make it compatible with the old code

    MemStorageSize!
    MemStorageStart!
  ;


// ////////////////////////////////////////////////////////////////////////////

public

  export  InitMemoryModule    ( dRomOffset nCount         --    ) // Initialize module selecting a user memory block
                                                                  // of maximally 64 kib

  export  MemErasePageSize    (                           -- q  ) // Size of a memory erase page
  export  MemWritePageSize    (                           -- q  ) // Size of a memory write page

  // Low level access everywhere

  export  MemEraseBlock       ( dRomOffs                  --    ) // Erase a memory block
  export  MemWriteBlock       ( nRamAddr dRomOffs nCount  --    ) // Write a block from RAM into memory
  export  MemReadBlock        ( nRamAddr dRomOffs nCount  --    ) // Read a block from memory to RAM
  export  MemVerifyBlock      ( nRamAddr dRomOffs nCount  -- f  ) // Verify a block of memory against RAM
  export  MemReadByte         ( dRomOffs                  -- b  ) // Read a byte    from memory
  export  MemReadWord         ( dRomOffs                  -- w  ) // Read a word    from memory
  export  MemReadTripple      ( dRomOffs                  -- t  ) // Read a tripple from memory
  export  MemReadQuad         ( dRomOffs                  -- q  ) // Read a quad    from memory


  // Reserved memory access (but no out of range address checks)

  export  MemStorageStart@    (                           -- d  ) // Get offset for user memory
  export  MemStorageStart!    ( d                         --    ) // Set offset for user memory

  export  MemUserEraseBlock   ( dRomOffs                  --    ) // Erase a user memory block
  export  MemUserWriteBlock   ( nRamAddr dRomOffs nCount  --    ) // Write a block from RAM into user memory
  export  MemUserReadBlock    ( nRamAddr dRomOffs nCount  --    ) // Read a block from user memory to RAM
  export  MemUserVerifyBlock  ( nRamAddr dRomOffs nCount  -- f  ) // Verify a block of user memory against RAM
  export  MemUserReadByte     ( dRomOffs                  -- b  ) // Read a byte    from user memory
  export  MemUserReadWord     ( dRomOffs                  -- w  ) // Read a word    from user memory
  export  MemUserReadTripple  ( dRomOffs                  -- t  ) // Read a tripple from user memory
  export  MemUserReadQuad     ( dRomOffs                  -- q  ) // Read a quad    from user memory

endModule

