
xVersion" 2011.11.5.1"

module Sets

  #document Sets

    // Defines small sets for x4th.
    //
    // Sets are 32 bit enities that reflect a set of things. A set can have up
    // to 32 elements. Sets can be created, elements can be added to it and
    // they can be removed. Sets can be added, subtracted and intersected.
    // A set can be tested to se if it contains some element.
    //
    // Elements are ordinal values 0 .. 31. Sets are bitmaps.
    //
    //
    //    Operator Operation     Operand_types  Result_type
    //
    //      [[[+]]]     union         set             set
    //      [[[-]]]     difference    set             set
    //      [[[*]]]     intersection  set             set
    //      [[[<=]]]    subset        set             flag
    //      [[[>=]]]    superset      set             flag
    //      [[[=]]]     equality      set             flag
    //      [[[<>]]]    inequality    set             flag
    //      [[[in]]]    membership    ordinal, set    flag
    //      [[>[]]]     make a set    ordinal         set
    //      [[[.]]]     empty set     -               set
    //
    //      [[\[\[]] Is a compiler defined word for creating set literals
    //      as in \\[\\[ 0 1 31 \\]\\] to be used like :
    //
    //        \\[\\[ 0 1 3 \\]\\] set setName
    //        : ...
    //          case
    //            \\[\\[ 0 1 3 \\]\\] in ... endIn
    //          endCase
    //          ...
    //        ;
    //
    //        : .... \\[\\[ 0 1 3 \\]\\] \\[\\[ 2 4 \\]\\] [+] .... ;
    //
    //
    // The following rules apply to [+], [-], and [*] :
    //
    // An ordinal b is in X Y [+] if and only if b is in X or Y (or both).
    //            b is in X Y [-] if and only if b is in X but not in Y.
    //            b is in X Y [*] if and only if b is in both X and Y.
    //
    // The following rules apply to [<=], [>=], [=], [<>], and [in] :
    //
    //   X Y [<=] is True just in case every member of X is a member of Y
    //   Z W [>=] is equivalent to W Z [<=]
    //   U V [=]  is True just in case U and V contain exactly the same
    //            members; otherwise, U V [<>] is True.
	  //   For an ordinal n and a set S, S n [in] is True just in case n is a
    //   member of S ( n is in S).

  #endDoc


private

  uses
    utils
  endUses


  // //////////////////////////////////////////////////////////////////////////

  alias dIndexToMask >[]    ( n           -- s )  #// Convert an ordinal into a set
  alias dor          [+]    ( s1 s2       -- s )  #// Union
  alias dand         [*]    ( s1 s2       -- s )  #// Intersection
  alias dnotAnd      [-]    ( s1 s2       -- s )  #// Difference s1 - s2, elements of s2 removed from s1
  alias d=           [=]    ( s1 s2       -- f )  #// Equality
  alias d0=          [.][=] ( s           -- f )  #// True when equal to empty set
  alias 2constant    set    ( s           --   )  #// Set constant constructor
  alias 2variable    setVar (             --   )  #// Set variable constructor
  alias d@           [@]    ( wRamAddr    -- s )  #// Set fetch
  alias d!           [!]    ( s wRamAddr  --   )  #// Set store

  [[ ]] set [.]                                   #// Empty set


  // //////////////////////////////////////////////////////////////////////////

  : [<=] ( s1 s2 -- f )

    #// True when s1 is a sub set of s2

    dOver dAnd [=]
  ;


  // //////////////////////////////////////////////////////////////////////////

  : [>=] ( s1 s2 -- f )

    #// True when s1 is a super set of s2

    dSwap [<=]
  ;


  // //////////////////////////////////////////////////////////////////////////

  : [<>] ( s1 s2 -- f )

    #// True when s1 not equal to s2

    [=] not
  ;


  // //////////////////////////////////////////////////////////////////////////

  : [in] ( s n -- f )

    #// membership, true when n is in s

    >[] dMaskCheck
  ;


  // //////////////////////////////////////////////////////////////////////////
  // //////////////////////////////////////////////////////////////////////////
  //
  // Add compiler support for in .. endIn, This can be used in case statements
  // like :
  //
  //   setVar s
  //
  //   [[ 3 7 ]] s [!]
  //
  //   [[ 8 12 13 21 ]] set t
  //
  //   n case
  //     [[ 1 2 31 ]] in ... endIn
  //     s            in ... endIn
  //     t            in ... endIn
  //   endCase
  //
  //
  //
  // The compiler wants to see three definitions
  //
  //   ((in))   ( n s -- n | n s -- ) // Helper for [[(in)]]
  //
  //     This can either be a code or a macro definition.
  //
  //     n is an integer that should be in the range 0 .. 31
  //     s is a set
  //     When n is in the set s the code between in and endIn should be
  //     executed, this should consume both n and s from the stack and
  //     on exit some value should be set such that (in) will know
  //     it should not branch or not (here the Zero flag in the status
  //     register will be cleared).
  //     When  n is not in s the code between in and endIn should not be
  //     executed and this should leave n on the stack for the next case
  //     to be evaluated. S should be consumed and some value should be
  //     returned such that (in) wil know it should branch (here the Zero
  //     flag in the status register will be set).
  //
  //   (in)     ( Label ) ( -- ) // Compiled [[in]]
  //
  //     This mist be a macro definition taking one parameter, the label
  //     to jump to when the [[in]] condition is not met. In this
  //     implementation the Zero flag in the status register is used, when
  //     set the branch will be taken to skip the part between in and endIn.
  //
  //     The iea is that ((in)) contains the bulk of the code and is
  //     implemented as a code word (so that it is called) and that (in)
  //     contains as little code as possible as it will be compiled inline
  //     (as a macro).
  //
  //   (endin)  ( Label1 Label2 ) ( -- ) // Compiled [[endIn]]
  //
  //     This must be a macro definition taking two parameters, the first
  //     one being a label to branch to after [[endIn]] and the other one
  //     is the label the [[in]] part jumps to when the condition is not
  //     met.
  //
  // These words should be added to the compiler vocaublary.

  also Compiler definitions


  code ((in)) ( n s -- n | n s -- ) // Helper for [[(in)]]

    PushLit   2                     ; // Perform 2 pick
                                    ; // (pick) ( ) ( u1 -- n )
                                    ; // Copy 2nd element to the top of the stack (0th :=: dup)
    add       w0, w0, w1
    neg       w1, w1                ; // Negate offset, stack goes up
    mov       [w9 + w1], w0
    rcall     <@" [in]">            ; // Check if n in s, returning ( f )
    PopReg    w1                    ; // Pop f into w1
    cp0       w1                    ; // Check if w1 == false, indicating n not in s
    btss      SR, #Z                ; // S/ n not in s
    mov       [w9--], w0            ; //    n in s, drop n (and skip later branch)
    return                          ; // B ----> done (Z flag when set indicates later branch is to be taken)
  endCode


  // //////////////////////////////////////////////////////////////////////////

  macro (in) ( Label ) ( -- ) // Compiled [[in]]
                                    ; // (in) - branch if Z flag set
    fBranch   $Label$
  endMacro


  // //////////////////////////////////////////////////////////////////////////

  macro (endin) ( Label1 Label2 ) ( -- ) // Compiled [[endIn]]
                                    ; // (endin) - unconditional branch
    bra       $Label1$

$Label2$:
  endMacro


  // //////////////////////////////////////////////////////////////////////////


public

  export set    ( s          --   ) // Set constant contructor
  export setVar (            --   ) // Set variable constructor
  export [@]    ( wRamAddr   -- s ) // Set fetch
  export [!]    ( s wRamAddr --   ) // Set store

  export >[]    ( n          -- s ) // Convert an ordinal into a set
  export [.]    (            -- s ) // Empty set
  export [.][=] ( s          -- f ) // True when s is empty set [.]

  export [+]    ( s s        -- s ) // Union
  export [*]    ( s s        -- s ) // Intersection
  export [-]    ( s s        -- s ) // Difference
  export [<=]   ( s s        -- f ) // Subset     check
  export [>=]   ( s s        -- f ) // Superset   check
  export [=]    ( s s        -- f ) // Equality   check
  export [<>]   ( s s        -- f ) // Inequality check
  export [in]   ( s n        -- f ) // Membership check

endModule

