
xVersion" 2012.10.10.1"

module Strings

  #document strings

  // String and memory words
  //
  //    Version 1, 2007-05-19
  //
  // All mentioned ANSI compatibilty holds only insofar that for this 4th CELL
  // size is non uniform; data CELLs are 16 bits and address CELLS are 24/32 bits.
  // ANSI requires CELL sizes to be uniform over all address spaces.
  //
  // A counted string is a string that is stored as a length byte followed by
  // the string data, counted strings are passed on the stack as a 16 bit
  // address only.
  //
  // A forth string is a string that is passed the stack as a 16 or 32 bit
  // pointer with an 8 bit length pushed over it. This three byte structure is
  // denoted by $<name> in stack pictures.
  //
  // Some words in this module operate on counted strings, others on forth
  // strings. The compiler word [[string]] compiles strings into ROM as counted
  // strings. Counted strings can be converted into 4th strings by using the
  // word [[RamCount]] (RomCount can't be easily supported for PIC24+)
  //
  // There is currently no support for string literals in target words.
  //
  // A lot of duplication is present in this module, the Harvard model really
  // shines here :p

  #endDoc

private

  uses
    fatal utils
  endUses


// ////////////////////////////////////////////////////////////////////////////

  $0000 constant Nil    #// Nil  pointer
  $0000 constant Null   #// Null pointer

  256 array $RomBuffer  #// Used fro ram / rom string comparison


// ////////////////////////////////////////////////////////////////////////////

  // Define string objects, able to hold a forth string as a length
  // and an address. Define some operations for 'm as well.

  {
    macro $ramvar ( aName ) ( -- )
    <!
      variable $aName$.len    // Holds string length
      variable $aName$.addr   // Holds string address

      : $aName$.len@   (      -- w        ) $aName$.len @                   ; // Get string length
      : $aName$.len!   ( w    --          ) $aName$.len !                   ; // Set string length
      : $aName$.len++  (      --          ) $aName$.len 1+!                 ; // Inc string length, no range checking
      : $aName$.len--  (      --          ) $aName$.len 1-!                 ; // Dec string length, no range checking
      : $aName$.len0=  (      -- f        ) $aName$.len@ 0=                 ; // Check string length to be zero
      : $aName$.len!0  (      -- f        ) $aName$.len0= not               ; // Check string length to be not zero

      : $aName$.addr@  (      -- w        ) $aName$.addr @                  ; // Get string address
      : $aName$.addr!  ( w    --          ) $aName$.addr !                  ; // Set string address
      : $aName$.addr++ (      --          ) $aName$.addr@ 1+ $aName$.addr!  ; // Inc string address, no range checking
      : $aName$.addr-- (      --          ) $aName$.addr@ 1- $aName$.addr!  ; // Dec string address, no range checking

      : $aName$!       ( $ram --          ) $aName$.len!  $aName$.addr!     ; // Store a string into a string var
      : $aName$@       (      -- $ram     ) $aName$.addr@ $aName$.len@      ; // Fetch a string from a string var
      : $aName$++      (      --          ) $aName$.len-- $aName$.addr++    ; // Advance in string like in "str  1 /string", no range checking
      : $aName$--      (      --          ) $aName$.len++ $aName$.addr--    ; // Go back in string like in "str -1 /string", no range checking
    >
    endMacro
    #document $ramvar
      //
      // Use $ramvar to define string variables that can hold a forth string. These are manipulater
      // and holding objects for forth strings.
      //
      // A forth string is transported over the stack as a 16 bit address topped by an 8 bit length.
      //
      // Use as in :
      //
      //   $ramvar aName
      //
      // to define a string variable. Two variables will be then be defined by the macro :
      //
      //   aName.len  an 8 bit variable that can hold a string length
      //   aName.addr a 16 bit variable that can hold a string address
      //
      // A couple of operations will be predefined on string variables, see macro text for details.
      //
      // The variables must be initilaized by placing a forth string into them before they can be used.
      // Initilaize as :
      //
      //    str-addr str-len aName!
      //
      // this will set up the address and the lenght part.
    #endDoc
  }

  // //////////////////////////////////////////////////////////////////////////

  code MemCompare ( nRamAddr1 nRamAddr2 nCount -- n ) // This is not ANSI compatible, see [[RamCompare]]

    ; // ///////////////////////////////////////////////////////////////////////
    ; //
    ; //        MemCompare ( Src1 Src2 Cnt -- n )
    ; //
    ; //        Src1 : 16 bits src1 address
    ; //        Src2 : 16 bits src2 address
    ; //        Cnt  : 16 bits byte count (block length)
    ; //
    ; //        Returns:
    ; //        -1 if Src1 < Src2
    ; //         0 if Src1 = Src2
    ; //         1 if Src1 > Src2
    ; //
    ; // Src1 and Src2 are indices that point to memory arrays
    ; // the lowest significant value must be in the lowest
    ; // memory location
    ; //
    ; // return value (n) :
    ; //
    ; //   -1 : Src1 < Src2
    ; //    0 : Src1 = Src2
    ; //    1 : Src1 > Src2
    ; //
    ; //  NOTE: Comparison start Src1 + Cnt - 1
    ; //
    ; // ///////////////////////////////////////////////////////////////////////

    PopReg    w3                    ; // Counter to w3
    PopReg    w2                    ; // Src2
                                    ; // Src1 in w0
    add       w0, w3, w0            ; // Point just past end of memory
    add       w2, w3, w2            ; // to be compared

MC_LP:

    dec       w3, w3                ; // Check for zero length
    bra       n, MC_SAME            ; // B/ Zero length strings are equal.

    mov.b     [--w0], w4            ; // Get src1 byte
    cp.b      w4, [--w2]            ; // Compare against src2 byte
    bra       eq, MC_LP             ; // B/ bytes are equal, test next bytes
    bra       lt, MC_LESS           ; // B/ src1 < src2

    retlw     #1, w0                ; // src1 > Src2

MC_SAME:

    retlw     #0, w0                ; // src1 == src2

MC_LESS:

    mov       #-1, w0               ; // src1 < src2
    return
  endCode
  #doc
    // MemCompare is ment for fixed length array comparisons where the lowest
    // significant byte is low in memory. So the compare must start at the
    // element of the array that is highest (last) in memory.
    //
    // Both arrays are in RAM space
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  code MemCopy ( nRamAddrSrc nRamAddrDst nCount -- ) #// Same as ANSI [[cmove]]

    ; // //////////////////////////////////////////////////////////////////////
    ; //
    ; //        MemCopy ( Src Dst Cnt -- )
    ; //
    ; //        Src : 16 bits src address in RAM
    ; //        Dst : 16 bits dst address in RAM
    ; //        Cnt : 16 bits byte count (block length)
    ; //
    ; //        Source and destination should not overlap
    ; //        and when they overlap only a copy to a lower
    ; //        address is possible (pointers go upwards)
    ; //
    ; //
    ; // //////////////////////////////////////////////////////////////////////

    PopReg    w2                    ; // Counter into w2
    PopReg    w1                    ; // Dst     into w1

    dec       w2, w2                ; // Dec count and check for zero count
    bra       n, MCP_NC             ; // B/ Nothing to copy

    repeat    w2                    ; // Repeat w2 + 1 times
    mov.b     [w0++], [w1++]

MCP_NC:

    drop
    return                          ; // B/ ----> done
  endCode


  // //////////////////////////////////////////////////////////////////////////

  code MemSet ( nRamAddr nVal nCount -- ) // non ANSI, also see [[Fill]] ( nRamAddr uCount bVal -- )

    ; // //////////////////////////////////////////////////////////////////////
    ; //
    ; //        MemSet ( Dst Val Cnt -- )
    ; //
    ; //        Dst : 16 bits dst address
    ; //        Val : 16 bits value to fill with
    ; //        Cnt : 16 bits byte count (block length)
    ; //
    ; // //////////////////////////////////////////////////////////////////////


    PopReg    w2                    ; // Pop nCount   into w2
    PopReg    w1                    ; // Pop nVal     into w1

    dec       w2, w2                ; // Check for zero bytes to clear
    bra       n, MS_NF              ; // B/ no fill

    repeat    w2                    ; // Repeat w2 + 1 times
    mov.b     w1, [w0++]            ; // Fill next char

MS_NF:

    drop                            ; // Remove nRamAddr from stack
    return                          ; // B/ ----> done

  endCode
  #doc
    // Arguments bVal and uCount are reversed from ANSI version (Fill)
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  code cmove> ( nRamAddrSrc nRamAddrDst bCount -- ) // ANSI compatible

    ; // //////////////////////////////////////////////////////////////////////
    ; //
    ; //        As cmove, but starting copy at high address
    ; //
    ; //        nRamAddrSrc : 16 bits src address in RAM
    ; //        nRamAddrDst : 16 bits dst address in RAM
    ; //        bCount      :  8 bits byte count (block length)
    ; //
    ; //        Source and destination should not overlap
    ; //        and when they overlap only a copy to a higher
    ; //        address is possible (pointers go downwards)
    ; //
    ; //
    ; // //////////////////////////////////////////////////////////////////////

    PopReg    w2                    ; // Counter into w2
    PopReg    w1                    ; // Dst     into w1

    dec       w2, w2                ; // Dec count and check for zero count
    bra       n, UCP_NC             ; // B/ Nothing to copy

    add       w0, w2, w0            ; // Point to end of src (last element)
    add       w1, w2, w1            ; // Point to end of dst (last element)

    repeat    w2                    ; // Repeat w2 + 1 times
    mov.b     [w0--], [w1--]

UCP_NC:

    Drop
    return                          ; // B/ ----> done
  endCode


  // //////////////////////////////////////////////////////////////////////////

  code RamPack ( $ram nRamAddr -- nRamAddr ) // Copy RAM string to RAM as a counted string

    PopReg    w2                    ; // nRamAddrDst into w2
    PopReg    w3                    ; // $ram.len into w3
    mov       w0, w1                ; // nRamAddrSrc into w1
    mov       w2, w0                ; // Keep a copy of nRamAddrDst in w0 (return value)

    mov.b     w3, [w2++]            ; // Move length to destination++
    dec       w3, w3                ; // Dec $ram.len and check for zero
    bra       n, RAP_done           ; // B/ Length was zero, done

    repeat    w3                    ; // Repeat w3 + 1 times
    mov.b     [w1++], [w2++]        ; // Move a byte

RAP_done:

    return                          ; // B / ----> done
  endCode


  // //////////////////////////////////////////////////////////////////////////

  code $RamCat ( $ram nRamAddr  -- ) // Concatenate $ram to counted string at nRamAddr

    PopReg    w1                    ; // Destination address into w1
    PopReg    w2                    ; // Source      length  into w2

    add       w2, [w1], [w1]        ; // Increment destination length with source length
    add       w2, w1, w1            ; // Add source length to dst address to get copy dst
    inc       w1, w1                ; // amd add one more

    dec       w2, w2                ; // Dec length and check for zero length
    bra       n, RAC_done           ; // B/ Zero length

    repeat    w2                    ; // Repeat w2 + 1 times
    mov.b     [w0++], [w1++]

RAC_done:

    drop
    return                          ; // B/ ----> done
  endCode


  // //////////////////////////////////////////////////////////////////////////

  code /ramstring ( $ram1 nCount -- $ram2 ) // remove nCount chars from front of $ram1 returning $ram2

    PopReg    w1                    ; // Get amount to remove into w1 - nCount
    PopReg    w2                    ; // Get length of $ram1  into w2 - nLen
    mov       w0, w3                ; // Get Ram address into w3

    cp        w1, w2                ; // Compare nCount - nLen
    bra       le, SSA_noclip        ; // B/ nCount <= nLen

    mov       w2, w1                ; // nCount > nLen, set nCount = nLen

SSA_noclip:

    sub       w2, w1, w2            ; // New length into w2
    add       w1, w3, w0            ; // Adjust pointer into w0
    PushReg   w2                    ; // Push new length

    return                          ; // B/ ----> done
  endCode


  // //////////////////////////////////////////////////////////////////////////

  code ramSkip ( $ram1 nChar -- $ram2 ) // skip leading n (char) in $ram1

    PopReg    w1                    ; // Char to match into w1
    and       #0xff, w1             ; // Keep low 8 bits
    PopReg    w2                    ; // $ram1.len into w2
    dec       w0, w0                ; // Point just before src (src ptr in w0)

RamSkip_loop:

    dec       w2, w2                ; // Check for zero remaining characters
    bra       n, RamSkip_done       ; // Reached end of $ram1 (remaining length was 0)

    cp.b      w1, [++w0]            ; // Comapre next byte
    bra       z, RamSkip_loop       ; // B/ match, not done

    ; // Fall through on no match

RamSkip_done:

    inc       w2, w2                ; // Bump length, we did a decrement too many
    PushReg   w2                    ; // New pointer is in w0, push length over it

    return                          ; // B/ ----> done
  endCode


  // //////////////////////////////////////////////////////////////////////////

  code -RamTrailChars ( $ram nChar -- $ram ) // Remove all trailing bChar characters from $ram


    PopReg    w1                    ; // nChar into w1
    and       #0xff, w1             ; // Keep low 8 bits
    PopReg    w2                    ; // nLen into w2, src ptr is in w0
    add       w0, w2, w3            ; // Point w3 just past end of string, keep w0 as original src

RATC_loop:

    dec       w2, w2                ; // Check for zero remaining length
    bra       n, RATC_done          ; // B/ length was zero, done

    cp.b      w1, [--w3]            ; // Check for match
    bra       z, RATC_loop          ; // B/ match, try next

RATC_done:

    inc       w2, w2                ; // Bump length, we did a decrement too many
    PushReg   w2                    ; // Push length over original src ptr (w0)
    return                          ; // B/ ----> done

  endCode


  // //////////////////////////////////////////////////////////////////////////

  code RamCompare ( $ram1 $ram2 -- b ) // compare RAMstring to RAMstring

    ; // //////////////////////////////////////////////////////////////////////
    ; //
    ; //      RamCompare ( $ram1 $ram2 -- n )
    ; //
    ; //        $ram1 : 16 bits RAM address, 16 bit length
    ; //        $ram2 : 16 bits RAM address, 16 bit length
    ; //
    ; //        Returns:
    ; //
    ; //        -1 if $ram1 < $ram2
    ; //         0 if $ram1 = $ram2
    ; //         1 if $ram1 > $ram2
    ; //
    ; //      This is a case sensitive lexicographic string compare
    ; //      i.e zz > za > z
    ; //
    ; // //////////////////////////////////////////////////////////////////////

    PopReg    w3                    ; // Length2 to w3
    PopReg    w2                    ; // Src2 to w2
    PopReg    w1                    ; // Length1 to w1
                                    ; // Src1 in w0
RC_LP:

    dec       w3, w3                ; // Check for zero length str2
    bra       n, RC_SAME1           ; // B/ Zero length strings are equal.

    dec       w1, w1                ; // Check for zero length str1
    bra       n, RC_SAME3           ; // B/ Zero length strings are equal.

    mov.b     [w0++], w4            ; // Get src1 byte
    cp.b      w4, [w2++]            ; // Compare against src2 byte
    bra       eq, RC_LP             ; // B/ bytes are equal, test next bytes
    bra       lt, RC_LESS           ; // B/ src1 < src2

RC_MORE:

    retlw     #1, w0                ; // src1 > Src2

RC_SAME1:

    dec       w1, w1                ; // w1 not decremented yet, do it now

RC_SAME3:

    cp        w1, w3                ; // Sofar strings equal, longest one is larger
    bra       lt, RC_LESS           ; // w1 < w3 -> len1 < len2 -> src1 < src2
    bra       gt, RC_MORE           ; // w1 > w2 -> len1 > len2 -> src1 > src2
    retlw     #0, w0                ; // otherwise strings equal length, src1 == src2

RC_LESS:

    mov       #-1, w0               ; // src1 < src2
    return
  endCode


  // //////////////////////////////////////////////////////////////////////////

  code StringBeginsWith ( dRomAddr nRamAddr -- f ) // True when nRamAddr starts with dRomAddr, both are counted stings

    ; // //////////////////////////////////////////////////////////////////////
    ; //
    ; //        StringBeginsWith ( dRomSrc wRamSrc -- f )
    ; //
    ; //        dRomAddr : 24 bits address in ROM space
    ; //        nRamAddr : 16 bits address in RAM space
    ; //
    ; //        Returns:
    ; //
    ; //          True  if nRamAddr^ starts with dRomAddr^
    ; //          False otherwise
    ; //
    ; // //////////////////////////////////////////////////////////////////////


    push      TBLPAG                ; // Don't change TBLPAG

    PopReg    w2                    ; // RAM address into w2
    PopReg    w1                    ; // High ROM address into w1
                                    ; // Low  ROM address is in w0
    mov       w1, TBLPAG            ; // TBLPAG = tblpage  ( rom)
    mov       w0, w5                ; // w5     = tbloffset( rom)

    tblrdl.b  [w5++], w3            ; // Copy rom length into w3
    and       #0xff, w3             ; // Keep low byte only
    cp.b      w3, [w2++]            ; // cp rom-len vs ram-len
    bra       gt, 3f                ; // B/ rom-len > ram-len, no possible match

    ; // ram-len >= rom-len, fall through to 1:
1:
    dec       w3, w3                ; // num--
    bra       n, 2f                 ; // B/ done, match

    tblrdl.b  [w5--], w4            ; // w4 = rom->hi-byte
    cp.b      w4, [w2++]            ; // rom->hi-byte = ram++^ ?
    bra       nz, 3f                ; // B/ done, no match

    dec       w3, w3                ; // num--
    bra       n, 2f                 ; // B/ done, match

    tblrdh.b  [w5], w4              ; // w4 = rom->upper-byte =
    cp.b      w4, [w2++]            ; // rom->upper-byte = ram++^?
    bra       nz, 3f                ; // B/ done, no match

    dec       w3, w3                ; // num--
    bra       n, 2f                 ; // B/ done, match

    add       w0, #2, w0            ; // src += 2
    addc      w1, #0, w1

    mov       w1, TBLPAG            ; // TBLPAG = tblpage  ( src)
    mov       w0, w5                ; // w5     = tbloffset( src)

    tblrdl.b  [w5++], w4            ; // w4 = rom->lo-byte
    cp.b      w4, [w2++]            ; // rom->lo-byte = ram++^ ?
    bra       z, 1b                 ; // B/ match, try next

    ; // no match, fall trhough to 3: to return false
3:
    mov       #0, w0                ; // No match, return false
    pop       TBLPAG                ; // Restore TBLPAG to entry value
    return                          ; // B/ ----> done

2:
    mov       #-1, w0               ; // A match, return true
    pop       TBLPAG                ; // Restore TBLPAG to entry value
    return                          ; // B/ ----> done
  endCode



  // //////////////////////////////////////////////////////////////////////////

  code Rom>RamCopy ( dRomAddrSrc nRamAddrDst uCount -- ) // Copy ROM data to RAM, uCount bytes

    ; // //////////////////////////////////////////////////////////////////////
    ; //
    ; //        Rom>RamCopy ( dRomSrc wRamDst bCnt -- )
    ; //
    ; //        Src : 32 bits src address in ROM - aligned on a 4 byte boundary
    ; //        Dst : 16 bits dst address in RAM
    ; //        Cnt : 16 bits byte count (block length)
    ; //
    ; // //////////////////////////////////////////////////////////////////////
    ; //
    ; //  Copy data from program memory to data memory
    ; //
    ; //  Registers used:  w0 w1 w2 w3 w4 w5
    ; //
    ; //  Inputs:
    ; //   w0, w1 = source address      (32 bits)
    ; //   w2     = destination address (16 bits)
    ; //   w3     = number of bytes (even or odd) - zero safe
    ; //
    ; // //////////////////////////////////////////////////////////////////////


    push      TBLPAG                ; // Don't change TBLPAG

    PopReg    w3                    ; // Length into w3
    PopReg    w2                    ; // Destination address in RAM into w2
    PopReg    w1                    ; // High ROM address into w1
                                    ; // Low  ROM address is in w0
1:
    dec       w3, w3                ; // num--
    bra       n, 2f                 ; // B/ done

    mov       w1, TBLPAG            ; // TBLPAG = tblpage  ( src)
    mov       w0, w5                ; // w5     = tbloffset( src)

    tblrdl.b  [w5++], [w2++]        ; // dst++ = lo byte

    dec       w3, w3                ; // num--
    bra       n, 2f                 ; // B/ done

    tblrdl.b  [w5--], [w2++]        ; // dst++ = hi byte

    dec       w3, w3                ; // num--
    bra       n, 2f                 ; // B/ done

    tblrdh.b  [w5], [w2++]          ; // dst++ = upper byte

    add       w0, #2, w0            ; // src += 2
    addc      w1, #0, w1

    bra       1b                    ; // B/ do next
2:
    drop                            ; // Loose last data stack item
    pop       TBLPAG                ; // Restore TBLPAG to entry value

    return                          ; // B/ ----> done

  endCode


  // //////////////////////////////////////////////////////////////////////////

  code Rom>RamStringCopy ( dRomAddrSrc nRamAddrDst -- ) // Copy a counted ROM string to RAM data. The RAM data is not counted

    ; // //////////////////////////////////////////////////////////////////////
    ; //
    ; //        Rom>RamStringCopy ( dRomSrc wRamDst -- )
    ; //
    ; //        Src : 32 bits src address in ROM - aligned on a 32 byte boundary
    ; //        Dst : 16 bits dst address in RAM
    ; //
    ; // //////////////////////////////////////////////////////////////////////

    push      TBLPAG                ; // Don't change TBLPAG

    PopReg    w2                    ; // Destination address in RAM into w2
    PopReg    w1                    ; // High ROM address into w1
                                    ; // Low  ROM address is in w0
    mov       w1, TBLPAG            ; // TBLPAG = tblpage  ( src)
    mov       w0, w5                ; // w5     = tbloffset( src)

    tblrdl.b  [w5++], w3            ; // Copy length into w3
    and       #0xff, w3             ; // Keep low byte only

1:
    dec       w3, w3                ; // num--
    bra       n, 2f                 ; // B/ done

    tblrdl.b  [w5--], [w2++]        ; // dst++ = hi byte

    dec       w3, w3                ; // num--
    bra       n, 2f                 ; // B/ done

    tblrdh.b  [w5], [w2++]          ; // dst++ = upper byte

    dec       w3, w3                ; // num--
    bra       n, 2f                 ; // B/ done

    add       w0, #2, w0            ; // src += 2
    addc      w1, #0, w1

    mov       w1, TBLPAG            ; // TBLPAG = tblpage  ( src)
    mov       w0, w5                ; // w5     = tbloffset( src)

    tblrdl.b  [w5++], [w2++]        ; // dst++ = lo byte

    bra       1b                    ; // B/ do next
2:
    drop                            ; // Loose last data stack item
    pop       TBLPAG                ; // Restore TBLPAG to entry value

    return                          ; // B/ ----> done
  endCode


  // //////////////////////////////////////////////////////////////////////////

  code Rom>RamStr ( dRomAddrSrc nRamAddrDst -- nRamAddrDst ) // Copy a counted ROM string to RAM data. The RAM data is counted

    ; // //////////////////////////////////////////////////////////////////////
    ; //
    ; //        Rom>RamStr ( dRomAddrSrc nRamAddrDst -- nRamAddrDst )
    ; //
    ; //        Src : 32 bits src address in ROM - aligned on a 32 byte boundary
    ; //        Dst : 16 bits dst address in RAM
    ; //
    ; // //////////////////////////////////////////////////////////////////////

    push      TBLPAG                ; // Don't change TBLPAG

    PopReg    w2                    ; // Destination address in RAM into w2
    PopReg    w1                    ; // High ROM address into w1
                                    ; // Low  ROM address is in w0
    mov       w1, TBLPAG            ; // TBLPAG = tblpage  ( src)
    mov       w0, w5                ; // w5     = tbloffset( src)

    push      w2                    ; // Keep a copy of original nRamAddrDst

    tblrdl.b  [w5++], w3            ; // Copy length into w3
    and       #0xff, w3             ; // Keep low byte only
    mov.b     w3, [w2++]            ; // Store length at destination

1:
    dec       w3, w3                ; // num--
    bra       n, 2f                 ; // B/ done

    tblrdl.b  [w5--], [w2++]        ; // dst++ = hi byte

    dec       w3, w3                ; // num--
    bra       n, 2f                 ; // B/ done

    tblrdh.b  [w5], [w2++]          ; // dst++ = upper byte

    dec       w3, w3                ; // num--
    bra       n, 2f                 ; // B/ done

    add       w0, #2, w0            ; // src += 2
    addc      w1, #0, w1

    mov       w1, TBLPAG            ; // TBLPAG = tblpage  ( src)
    mov       w0, w5                ; // w5     = tbloffset( src)

    tblrdl.b  [w5++], [w2++]        ; // dst++ = lo byte

    bra       1b                    ; // B/ do next
2:
    pop       w0                    ; // Return original nRamAddrDst

    pop       TBLPAG                ; // Restore TBLPAG to entry value

    return                          ; // B/ ----> done
  endCode


  // //////////////////////////////////////////////////////////////////////////

  code rom>$ram ( dRomAddrSrc nRamAddrDst -- $ram ) // Copy a counted ROM string to a ram string

    ; // //////////////////////////////////////////////////////////////////////
    ; //
    ; //        rom>$ram ( dRomAddrSrc nRamAddrDst -- $ram )
    ; //
    ; //        Src : 32 bits src address in ROM - aligned on a 32 byte boundary
    ; //        Dst : 16 bits dst address in RAM
    ; //
    ; // //////////////////////////////////////////////////////////////////////

    push      TBLPAG                ; // Don't change TBLPAG

    PopReg    w2                    ; // Destination address in RAM into w2
    PopReg    w1                    ; // High ROM address into w1
                                    ; // Low  ROM address is in w0
    mov       w1, TBLPAG            ; // TBLPAG = tblpage  ( src)
    mov       w0, w5                ; // w5     = tbloffset( src)

    tblrdl.b  [w5++], w3            ; // Copy length into w3
    and       #0xff, w3             ; // Keep low byte only

    push      w3                    ; // Push copy of original rom length
    push      w2                    ; // Push copy of original ram dst address

1:
    dec       w3, w3                ; // len--
    bra       n, 2f                 ; // B/ done

    tblrdl.b  [w5--], [w2++]        ; // dst++ = hi byte

    dec       w3, w3                ; // len--
    bra       n, 2f                 ; // B/ done

    tblrdh.b  [w5], [w2++]          ; // dst++ = upper byte

    dec       w3, w3                ; // len--
    bra       n, 2f                 ; // B/ done

    add       w0, #2, w0            ; // src += 2
    addc      w1, #0, w1

    mov       w1, TBLPAG            ; // TBLPAG = tblpage  ( src)
    mov       w0, w5                ; // w5     = tbloffset( src)

    tblrdl.b  [w5++], [w2++]        ; // dst++ = lo byte

    bra       1b                    ; // B/ do next
2:
    pop       w0                    ; // Pop original ram dst address
    pop       w1                    ; // Pop original rom length
    PushReg   w1                    ; // Return $ram ( pointer length )

    pop       TBLPAG                ; // Restore TBLPAG to entry value

    return                          ; // B/ ----> done
  endCode


  // //////////////////////////////////////////////////////////////////////////

  code RomStringCompare ( dRomAddrSrc nRamAddrSrc -- nResult ) // Non ANSI, this is for for counted strings.

    ; // //////////////////////////////////////////////////////////////////////
    ; //
    ; //        RomStringCompare ( dRomSrc wRamSrc -- bResult )
    ; //
    ; //        dRomSrc : 32 bits address in ROM space
    ; //        wRamSrc : 16 bits address in RAM space
    ; //
    ; //        Returns:
    ; //
    ; //          -1 if RAM < ROM
    ; //           0 if RAM = ROM
    ; //           1 if RAM > ROM
    ; //
    ; //        According to 'standard' lexicographical ordering
    ; //
    ; // //////////////////////////////////////////////////////////////////////


    push      TBLPAG                ; // Don't change TBLPAG

    PopReg    w2                    ; // RAM address into w2
    PopReg    w1                    ; // High ROM address into w1
                                    ; // Low  ROM address is in w0
    mov       w1, TBLPAG            ; // TBLPAG = tblpage  ( rom)
    mov       w0, w5                ; // w5     = tbloffset( rom)

    tblrdl.b  [w5++], w3            ; // Copy rom length into w3
    and       #0xff, w3             ; // Keep low byte only
    mov.b     [w2++], w6            ; // Copy ram length into w6
    and       #0xff, w6             ; // Keep low byte only

    ; // Fall through to 1: , enter loop
1:
    dec       w3, w3                ; // Check for zero rom-len
    bra       n, RRC_SAME3          ; // B/ Zero length strings are still equal.
    dec       w6, w6                ; // Check for zero ram-len
    bra       n, RRC_SAME6          ; // B/ Zero length strings are still equal.

    tblrdl.b  [w5--], w4            ; // w4 = rom->hi-byte
    cp.b      w4, [w2++]            ; // rom->hi-byte - ram++^
    bra       lt, RRC_MORE          ; // B/ ram > rom
    bra       gt, RRC_LESS          ; // B/ ram < rom

    dec       w3, w3                ; // Check for zero rom-len
    bra       n, RRC_SAME3          ; // B/ Zero length strings are still equal.
    dec       w6, w6                ; // Check for zero ram-len
    bra       n, RRC_SAME6          ; // B/ Zero length strings are still equal.

    tblrdh.b  [w5], w4              ; // w4 = rom->upper-byte =
    cp.b      w4, [w2++]            ; // rom->upper-byte = ram++^?
    bra       lt, RRC_MORE          ; // B/ ram > rom
    bra       gt, RRC_LESS          ; // B/ ram < rom

    dec       w3, w3                ; // Check for zero rom-len
    bra       n, RRC_SAME3          ; // B/ Zero length strings are still equal.
    dec       w6, w6                ; // Check for zero ram-len
    bra       n, RRC_SAME6          ; // B/ Zero length strings are still equal.

    add       w0, #2, w0            ; // src += 2
    addc      w1, #0, w1

    mov       w1, TBLPAG            ; // TBLPAG = tblpage  ( src)
    mov       w0, w5                ; // w5     = tbloffset( src)

    tblrdl.b  [w5++], w4            ; // w4 = rom->lo-byte
    cp.b      w4, [w2++]            ; // rom->lo-byte = ram++^ ?
    bra       lt, RRC_MORE          ; // B/ ram > rom
    bra       gt, RRC_LESS          ; // B/ ram < rom

    bra       1b                    ; // Try next byte

    ; // FALL ?

RRC_SAME3:

    dec       w6, w6                ; // w1 not decremented yet, do it now

RRC_SAME6:

    cp        w6, w3                ; // Sofar strings equal, longest one is larger
    bra       lt, RRC_LESS          ; // w6 < w3 -> len < len -> src < src
    bra       gt, RRC_MORE          ; // w6 > w3 -> len1 > len2 -> src1 > src2

    ; // otherwise strings equal length, src1 == src2, fall trhough to RRC_EQUAL

RRC_EQUAL:

    clr       w0                    ; // RAM = ROM
    bra       RRC_EXIT              ; // B/ done

RRC_MORE:

    mov       #1, w0                ; // RAM > ROM
    bra       RRC_EXIT              ; // B/ done

RRC_LESS:

    mov       #-1, w0               ; // RAM < ROM

    ; // Fall through to RRC_EXIT

RRC_EXIT:
    pop       TBLPAG                ; // Restore TBLPAG

    return                          ; // B/ ----> done
  endCode


  // //////////////////////////////////////////////////////////////////////////

  alias 0= Nil=       ( nAddr -- f ) #// True when wAddr = Nil
  alias 0= Null=      ( nAddr -- f ) #// True when wAddr = Null


  // //////////////////////////////////////////////////////////////////////////

  : ?Nil<> ( nAddr -- nAddr fTrue | nAddr -- fFalse )

    #// Returns False if nAdsdr is Nil
    #// otherwise it returns the nAddr with True over it.
    #//
    #// This one used to be called ?ptrDup which is a misleading name

    aDup Nil=
    if
      aDrop False
    else
      True
    endIf
  ;
    noStackCheck

  // //////////////////////////////////////////////////////////////////////////

  alias aDup@ RamStrLen ( nRamAddr -- nRamAddr bLength ) #// Get length for counted RAM string,
  #doc
    // also see [[RamCount]]
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  : RamStrEmpty? ( nRamAddr -- fTrue | nRamAddr -- nRamAddr fFalse )

    #// For counted strings

    RamStrLen 0=
    if
      aDrop
      True
    else
      False
    endIf
  ;
    noStackCheck

  // //////////////////////////////////////////////////////////////////////////

  : RamStrNotEmpty? ( nRamAddr -- nRamAddr fTrue | nRamAddr -- fFalse )

    #// For counted strings

    RamStrLen 0=
    if
      aDrop
      False
    else
      True
    endIf
  ;


  // //////////////////////////////////////////////////////////////////////////
  // New string words, borrowed from ANSI/ISO and Open Firmware
  // sources.
  //
  //    $    : wAddr bLen
  //    $ram : nRamAddr bLen
  //
  // //////////////////////////////////////////////////////////////////////////


  code Fill ( nRamAddr nCount nVal -- ) // ANSI compatible Fill, also see [[MemSet]] ( nRamAddr bVal uCount )

    PopReg    w1                    ; // Pop nVal     into w1
    PopReg    w2                    ; // Pop nCount   into w2

    dec       w2, w2                ; // Check for zero bytes to clear
    bra       n, F_NF               ; // B/ no fill

    repeat    w2                    ; // Repeat w2 + 1 times
    mov.b     w1, [w0++]            ; // Fill next char

F_NF:

    drop                            ; // Remove nRamAddr from stack
    return                          ; // B/ ----> done
  endCode


  // //////////////////////////////////////////////////////////////////////////

  : Erase ( nRamAddr uCount -- )

    #// zero fill

    0 Fill
  ;


  // //////////////////////////////////////////////////////////////////////////

  : Blank ( nRamAddr uCount -- )

    #// fill with spaces

    bl Fill
  ;


  // //////////////////////////////////////////////////////////////////////////

  alias MemCopy cmove ( nRamAddrSrc nRamAddrDst bCount -- ) #// ANSI compatible


  // //////////////////////////////////////////////////////////////////////////

  : RamPlace ( $ram nRamAddr -- )

    #// Copy ROM string to RAM as a counted string

    RamPack
    aDrop
  ;

  // //////////////////////////////////////////////////////////////////////////

  code RamCount ( nRamAddr -- $ram ) // Count a RAM string

    mov.b     [w0++], w1
    and       #0xff, w1
    PushReg   w1
    return
  endCode


  // //////////////////////////////////////////////////////////////////////////

  : $RamSave ( $ram nRamAddr -- $ram )

    #// Save string as counted, then count it.

    RamPack
    RamCount
  ;


  // //////////////////////////////////////////////////////////////////////////

  : -Ramtrailing ( $ram -- $ram )

    #// remove trailing blanks from $ram

    bl -RamtrailChars
  ;


  // //////////////////////////////////////////////////////////////////////////

  : -RamNulls ( $ram -- $ram )

    #// remove trailing nulls from $ram

    0 -RamtrailChars
  ;


  // //////////////////////////////////////////////////////////////////////////

  : RomString=Ram ( dRomAddrSrc nRamAddrSrc -- f )

    #// For counted strings

    RomStringCompare 0=
  ;


  // //////////////////////////////////////////////////////////////////////////

  alias SwapDrop  "clip"    ( $str1 n     -- $str2       ) #// Set a new length for $str, can be RAM or ROM string
  alias ddrop     $ramdrop  ( $ram        --             ) #// Drop a forth RAM string
  alias ddup      $ramdup   ( $ram        -- $ram $ram   ) #// Dup  a forth RAM string
  alias dswap     $ramswap  ( $ram1 $ram2 -- $ram2 $ram1 ) #// Swap two forth RAM strings


  // //////////////////////////////////////////////////////////////////////////


  // Define some helper objects for the string search words

  $ramvar haystack
  $ramvar haystack.orig
  $ramvar needle


  // //////////////////////////////////////////////////////////////////////////

  : Search ( $ram1 $ram2 -- $ram3 f )

    #// Search $ram1 in $ram2 according to ANSI specs
    #//
    #// This is a straight forward naive string search implementation
    #//
    #// Search the string specified by $xxx1 for the string specified by $xxx2.
    #// If f is true, a match was found at $xxx3 with the length of $xxx3
    #// characters remaining.
    #// If flag is false there was no match and $xxx3 is $xxx1.
    #//
    #// $xxx1 is the haystack, $xxx2 is the needle.

    $ramdup haystack!
    haystack.orig!                  // Keep two copies of refs to haystack
    needle!                         // keep a copy of a ref to the needle
    needle.len0=                    // When needle.len = 0 we have a match
    if
      True                          // Flag match
    else                            // needle.len > 0
      haystack.len@
      needle.len@ -                 // Check if needle <= haystack
      dup 0<
      if
        DropFalse                   // Needle > haystack, no possible match
      else                          // ( outer_loop_count  - 1) - number of compares to make
        1+ 0 swap
        do
          needle@ haystack@         // Check if needle at current haystack position
          needle.len@ "clip"
          RamCompare 0=
          if                        // We have a match
            True                    // Leave a true flag
            leave                   // Leave the do loop
          else
            haystack.len0=          // Through the haystack ?
            if                      // yes
              False                 // Leave a False flag
              leave                 // Leave the do loop
            else
              haystack++            // Try further on in haystack
            endIf
          endif
        loop
      endif
    endif
    if                              // Match found
      haystack.addr@                //   return altered haystack
      needle.len@ +
      haystack.len@
      needle.len@ -
      True                          //   and flag match
    else                            // No match found
      haystack.orig@                //   return original haystack
      False                         //   and flag no match
    endif
  ;


  // //////////////////////////////////////////////////////////////////////////

  : RomSearch ( $ram1 dRomAddress -- $ram3 f )

    #// Search counted string at dRomAddress in $ram1

    $RomBuffer rom>$ram             // ( $ram1 $ram2 )
    Search
  ;


  // //////////////////////////////////////////////////////////////////////////

  : lower? ( b -- f )

    #// test b is a lower case character

    'a 'z InRange
  ;


  // //////////////////////////////////////////////////////////////////////////

  : upper? ( b -- f )

    #// test b is a upper case character

    'A 'Z InRange
  ;


  // //////////////////////////////////////////////////////////////////////////

  : >upper ( b -- b )

    #// make b into it's upper case variant

    dup lower?
    bl and xor
  ;


  // //////////////////////////////////////////////////////////////////////////

  : >lower ( b -- b )

    #// make b into it's lower case variant

    dup upper?
    bl and xor
  ;


  // //////////////////////////////////////////////////////////////////////////

  : $upper ( $ram -- )

    #// make $ram uppercase

    0 ?do                           // If non zero length, do work
      aDupc@                        // Dup address, get char
      >upper                        // ( nAddr C )
      over c!                       // ( nAddr )
      1a+                           // ( nAddr )
    loop
    aDrop                           // ( )
  ;


  // //////////////////////////////////////////////////////////////////////////

  : $lower ( $ram -- )

    #// make $ram uppercase

    0 ?do                           // If non zero length, do work
      aDupc@                        // Dup address, get char
      >lower                        // ( nAddr c )
      over c!                       // ( nAddr )
      1a+                           // ( nAddr )
    loop
    aDrop                           // ( )
  ;


  // //////////////////////////////////////////////////////////////////////////

public


  // //////////////////////////////////////////////////////////////////////////
  //
  // Old string words, present for backward compatibility
  // Some names have changed though

  export  Nil               (                                 -- wNil                     ) // General Nil pointer, alias for Null
  export  Null              (                                 -- wNil                     ) // General Null pointer, alias for Nil
  export  MemCompare        ( nRamAddr1 nRamAddr2 bCount      -- f                        ) // This is not ANSI compatible, see [[RamCompare]]
  export  MemCopy           ( nRamAddrSrc nRamAddrDst bCount  --                          ) // Same as ANSI [[cmove]]
  export  MemSet            ( nRamAddr bVal uCount            --                          ) // non ANSI, also see [[Fill]] ( nRamAddr uCount bVal -- )
  export  Nil=              ( wAddr                           -- f                        ) // True when wAddr = Nil
  export  Null=             ( wAddr                           -- f                        ) // True when wAddr = Nil
  export  ?nil<>            ( wAddr                           -- wAddr fTrue | fFalse     ) // Returns just false if wAddr = Nil, wAddr True otherwise
  export  RamStrLen         ( nRamAddr                        -- nRamAddr bLength         ) // Get length for counted RAM string,
  export  RamStrEmpty?      ( nRamAddr                        -- fTrue | wRamAdrr fFalse  ) // For counted strings
  export  RamStrNotEmpty?   ( nRamAddr                        -- nRamAddr fTrue | fFalse  ) // For counted strings


  // New string words borrowed from ANSI/ISO and Open firmware. and some utilities

  // $ram :~: nRamAddr bLen

  export  $ramvar           (                                 --                          ) // Define a pseudo object holding forth strings (wAddr bLen ) sets
  export  Fill              ( nRamAddr uCount bVal            --                          ) // ANSI compatible Fill, also see MemSet ( nRamAddr bVal uCount )
  export  Erase             ( nRamAddr uCount                 --                          ) // zero fill)
  export  Blank             ( nRamAddr uCount                 --                          ) // fill with spaces
  export  cmove             ( nRamAddrSrc nRamAddrDst bCount  --                          ) // Same as MemCopy, tested, ANSI compatible
  export  cmove>            ( nRamAddrSrc nRamAddrDst bCount  --                          ) // ANSI compatible
  export  RamPack           ( $ram nRamAddr                   -- nRamAddr                 ) // Copy RAM string to RAM as a counted string
  export  RamPlace          ( $ram nRamAddr                   --                          ) // Copy RAM string to RAM as a counted string
  export  RamCount          ( nRamAddr                        -- $ram                     ) // Count a ROM string, returning a forth RAM string
  export  $RamSave          ( $ram nRamAddr                   -- $ram                     ) // Save RAM string as counted in RAM, then count it.
  export  $RamCat           ( $ram nRamAddr                   --                          ) // Concatenate $ram to counted string at nRamAddr
  export  /ramstring        ( $ram1 bCount                    -- $ram2                    ) // remove bCount chars from $ram1 returning $ram2,
  export  ramSkip           ( $ram1 b                         -- $ram2                    ) // skip leading b in $ram1
  export  "clip"            ( $str b                          -- $str                     ) // Set a new length for $str, can be RAM or ROM string
  export  -RamTrailChars    ( $ram bChar                      -- $ram                     ) // Remove trailing bChar from $ram
  export  -Ramtrailing      ( $ram                            -- $ram                     ) // remove trailing blanks from $ram
  export  -RamNulls         ( $ram                            -- $ram                     ) // remove trailing nulls from $ram
  export  RamCompare        ( $ram1 $ram2                     -- b                        ) // compare RAMstring1 to RAMstring2
  export  $ramdrop          ( $ram                            --                          ) // Drop a forth RAM string
  export  $ramdup           ( $ram                            -- $ram $ram                ) // Dup  a forth RAM string
  export  $ramswap          ( $ram1 $ram2                     -- $ram2 $ram1              ) // Swap two forth RAM strings

  export  Search            ( $ram1 $ram2                     -- $ram3 f                  ) // Search $ram2 in $ram1 according to ANSI specs
  export  lower?            ( b                               -- f                        ) // test b is a lower case character
  export  >lower            ( b                               -- b                        ) // make b into it's lower case variant
  export  $lower            ( $ram                            --                          ) // make $ram lower case
  export  upper?            ( b                               -- f                        ) // test b is a upper case character
  export  >upper            ( b                               -- b                        ) // make b into it's upper case variant
  export  $upper            ( $ram                            --                          ) // make $ram upper case

  // ROM stuff

  export  Rom>RamCopy       ( dRomAddrSrc nRamAddrDst uCount  --                          ) // Copy ROM data to RAM, uCount bytes
  export  Rom>RamStringCopy ( dRomAddrSrc nRamAddrDst         --                          ) // Copy a counted ROM string to RAM data. The RAM data is not counted
  export  StringBeginsWith  ( dRomAddrSrc nRamAddrSrc         -- f                        ) // True when nRamAddr starts with dRomAddr, both are counted stings
  export  RomStringCompare  ( dRomAddrSrc nRamAddrSrc         -- bResult                  ) // Non ANSI, this for for counted strings.
  export  RomString=Ram     ( dRomAddrSrc nRamAddrSrc         -- f                        ) // For counted strings
  export  Rom>RamStr        ( dRomAddrSrc nRamAddrDst         -- nRamAddrDst              ) // Copy a counted ROM string to RAM data. The RAM data is counted
  export  rom>$ram          ( dRomAddrSrc nRamAddrDst         -- $ram                     ) // Copy a counted ROM string to a ram string
  export  RomSearch         ( $ram1 dRomAddress               -- $ram3 f                  ) // Search counted sring at dRomAddress in $ram1

endModule

