
xVersion" 2012.10.10.1"

module utils

  #document utils

    // Utility words
    //
    // About ROM byte addressing:
    //
    //   A special set of words has been been made to allow for arbitrary ROM
    //   byte access. As the processor has no direct byte addressing in ROM
    //   but long word addressing instead this is a bit kludgy ...
    //
    //   ... however once a long word is pointed to there are special
    //   registers tat allow for byte access, these are tblptrl and tblptrh.
    //   Tblptrl can be used to access the lower two bytes and tblptrh to
    //   access the upper byte.
    //
    //   ROM addressing uses a special 6 byte pointer consisting of a four
    //   byte base address and a two byte byte-index, the latter should only
    //   hold values [0..2] to point to the low, high and upper bytes
    //   of the 24 bit word addressed by the base address. The C compiler
    //   and the assembler will store ROM data starting at an even ROM
    //   address (as there are no odd ROM addresses really, the pointer is
    //   a word (16 bits) address pointer but the ROM is arranged as long
    //   words (32 bits of which 24 only are used).
    //
    // ROM addressing glossary :
    //
    //   RomPtr! ( dRomAddr -- )
    //
    //     Sets up the ROM pointer to point to the low byte at dRomAddr
    //     this is used to initialize the ROM pointer such that the
    //     following words can be used. dRomAddr should be an even
    //     ROM address as obtained with the #tbloffset and #tablepage
    //     assembler macros.
    //
    //   RomPtr++ ( -- )
    //
    //     Advance the ROM pointer by 1 by incrementing the byte index,
    //     when the byte index goes > 2 the byte index is cleared to 0
    //     and the dAddrr bit is incremented by 2 to point to the next
    //     dWord in ROM.
    //
    //   RomPtr-- ( -- )
    //
    //     Advance the ROM pointer by -1 by decrementing the byte index,
    //     when the byte index goes < 0 the byte index is cleared to 2
    //     and the dAddrr bit is decremented by 2 to point to the prev.
    //     dWord in ROM.
    //
    //   romc@ ( -- n )
    //
    //     Read a byte from the location pointed to by the ROM pointer
    //
    //   romc! ( n -- )
    //
    //     Store n as a byte at the location pointed to by the ROM pointer
    //
    //   romc@++ ( -- n )
    //
    //     As romc@, post increment the ROM pointer
    //
    //   romc!++ ( n -- )
    //
    //     As romc!, post increment the ROM pointer
    //
    //   rom++c@ ( -- n )
    //
    //     As romc@, pre increment the ROM pointer
    //
    //   rom++c! ( n -- )
    //
    //     As romc!, pre increment the ROM pointer
    //
    //   romc@-- ( -- n )
    //
    //     As romc@, post decrement the ROM pointer
    //
    //   romc!-- ( n -- )
    //
    //     As romc!, post decrement the ROM pointer
    //
    //   rom--c@ ( -- n )
    //
    //     As romc@, pre decrement the ROM pointer
    //
    //   rom--c! ( n -- )
    //
    //     As romc!, pre decrement the ROM pointer
    //
    //

  #endDoc


private


  // //////////////////////////////////////////////////////////////////////////

  variable CRC            #// Crc8 and crc16 result variable.   , and General Purpose
 3variable _romPtr        #// Rom pointer (d) + byte offset (n), 6 bytes in total
                          #// low word has byte offest
                          #// upper two words hold a ROM word address pointer
                          #// This is not the ordinary ROM pointer which is 4 bytes
                          #// in size.

  varInterface     CRC
  bytePtr          ram    #// Define ram(Ptr) as a bytePtr


  // //////////////////////////////////////////////////////////////////////////

  code ud2* ( d -- d )                // Multiply d by 2, unsigned

    ; // ab * cd -> a*c*2^32 + (ad + bc)*2^16 + bd
    ; // with b = 2 and a = 0 and ignoring the 2^32 term
    ; // 2c*2^16 + 2d

    PopReg    w1                    ; // high d = c in w1
                                    ; // low  d = d in w0
    mul.uu    w1, #2, w2            ; // 2c          -> w3:w2
    mul.uu    w0, #2, w0            ; // 2d          -> w1:w0
    add       w1, w2, w1            ; // 2d.h + 2c.l -> w1
    PushReg   w1                    ; // Push high result, low was in w0

    return                          ; // B/ ----> done
  endCode

  // //////////////////////////////////////////////////////////////////////////
  //
  // Rom byte access words, a special 6 byte ROM pointer is used for this,
  // see module documentation

  code rom_c@ ( nByte dWordAddr -- n )


    push      TBLPAG                ; // Don't change TBLPAG

    PopReg    TBLPAG                ; // Get high dWordAddress into TBLPAG
    PopReg    w1                    ; // Get low  dWordAddress into w1

    and       #3, w0                ; // w0 should be in [0..2], accept 3 with incorrect results but no crash
    jr_f      w0                    ; // Dispatch on byte address [0..2]

    bra       0f                    ; // low   byte
    bra       0f                    ; // high  byte
                                    ; // upper byte
    tblrdh.b  [w1], w0              ; // w0 = upper byte
    bra       1f                    ; // B/ done

0:
    add       w1, w0, w1            ; // Address low or high byte depending on w1
    tblrdl.b  [w1], w0              ; // w0 = low byte or w0 = high byte

1:
    and       #0xff, w0             ; // Keep low byte only
    pop       TBLPAG                ; // Restore TBLPAG to entry value
    return                          ; // B/ ----> done
  endCode


  // //////////////////////////////////////////////////////////////////////////

  code rom_c! ( n nByte dWordAddr -- )


    push      TBLPAG                ; // Don't change TBLPAG

    PopReg    TBLPAG                ; // Get high dWordAddress into TBLPAG
    PopReg    w1                    ; // Get low  dWordAddress into w1
    PopReg    w2                    ; // Get byte to address [0..2]
    and       #3, w2                ; // Accept 3 for w2, giving incorrect results but no crash

    mov       w0, w3                ; // n into w3
    jr_f      w2                    ; // Dispatch on byte address

    bra       0f                    ; // low   byte
    bra       0f                    ; // high  byte
                                    ; // upper byte
    tblwth.b  w3, [w1]              ; // dst.upper_byte = w3
    bra       1f                    ; // B/ done

0:
    add       w1, w2, w1            ; // Address low or high byte depending on w1
    tblwtl.b  w3, [w1]              ; // dst.low_byte = w3 or dst.high_byte = w3

1:
    drop                            ; // Drop last stack item
    pop       TBLPAG                ; // Restore TBLPAG to entry value
    return                          ; // B/ ----> done
  endCode


  // //////////////////////////////////////////////////////////////////////////

  code romPtr! ( dRomAddr -- )

    ; // _RomPtr 0!                   // Clear byte index
    ; // _RomPtr 2+ d!                // Store dRomAddr
                                    ; // Clear byte index
    mov       #<@" _romPtr">, w1
    clr       [w1]
    add       #4, w1                ; // Pop dRomAddr into address field
    mov       w0, [w1--]
    mov       [w9--], [w1]
    drop                            ; // Drop
    return                          ; // B/ ----> done
  endCode

  // //////////////////////////////////////////////////////////////////////////

  code romPtr@ ( -- t )

    ; // _romPtr t@                   // Get word address and byte offset

    mov       #<@" _romPtr">, w1
    mov       w0, [++w9]
    mov       [w1++], [++w9]
    mov       [w1++], [++w9]
    mov       [w1++], w0

    return

  endCode


  // //////////////////////////////////////////////////////////////////////////

  code romPtr++ ( -- )
    mov       #<@" _romPtr">, w1    ; // Point w1 to byte index
    mov       #3, w2                ; // work modulo 3
    inc       [w1], [w1]            ; // Increment byte index
    cp        w2, [w1]              ; // Check for wrap modulo 3
    bra       gt, 1f                ; // B/ no wrap

    clr       w2                    ; // Make a zero in w2
    clr       [w1++]                ; // Set byte index to 0, point w1 to low  dAddr
    inc2      [w1], [w1++]          ; // Increment low dAddr, point w1 to high dAddr
    addc      w2, [w1], [w1]        ; // Pass carry to high dAddr

1:
    return                          ; // B/ ----> done
  endCode

  // //////////////////////////////////////////////////////////////////////////

  code romPtr-- ( -- )
    mov       #<@" _romPtr">, w1    ; // Point w1 to byte index
    dec       [w1], [w1]            ; // Decrement byte index
    bra       nn, 1f                ; // When not negative, done

    mov       #2, w2
    mov       w2, [w1++]            ; // Set byte index to 2, point w1 to low  dAddr
    clr       w2                    ; // Make a zero in w2
    dec2      [w1], [w1++]          ; // Decrement low dAddr, point w1 to high dAddr
    subb      w2, [w1], [w1]        ; // Pass borrow to high dAddr

1:
    return                          ; // B/ ----> done
  endCode

  // //////////////////////////////////////////////////////////////////////////

  : romc@ ( -- n )

    romPtr@ rom_c@
  ;


  // //////////////////////////////////////////////////////////////////////////

  : romc! ( n -- )

    romPtr@ rom_c!
  ;


  // //////////////////////////////////////////////////////////////////////////

  : romc@++ ( -- n )

    romPtr@ rom_c@
    romPtr++
  ;


  // //////////////////////////////////////////////////////////////////////////

  : romc!++ ( n -- )

    romPtr@ rom_c!
    romPtr++
  ;


  // //////////////////////////////////////////////////////////////////////////

  : rom++c@ ( -- n )

    romPtr++
    romPtr@ rom_c@
  ;


  // //////////////////////////////////////////////////////////////////////////

  : rom++c! ( n -- )

    romPtr++
    romPtr@ rom_c!
  ;


  // //////////////////////////////////////////////////////////////////////////

  : romc@-- ( -- n )

    romPtr@ rom_c@
    romPtr--
  ;


  // //////////////////////////////////////////////////////////////////////////

  : romc!-- ( n -- )

    romPtr@ rom_c!
    romPtr--
  ;


  // //////////////////////////////////////////////////////////////////////////

  : rom--c@ ( -- n )

    romPtr--
    romPtr@ rom_c@
  ;


  // //////////////////////////////////////////////////////////////////////////

  : rom--c! ( n -- )

    romPtr--
    romPtr@ rom_c!
  ;


  // //////////////////////////////////////////////////////////////////////////

  code inRange ( u1 u2 u3 -- f ) // True when u2 <= u1 <= u3

    PopReg    w3
    PopReg    w2
    mov       w0, w1
    mov       #False, w0            ; // Assume not in range

    sub       w3, w1, w3            ; // u3 - u1 -> u3

    btss      SR, #C                ; // S/ carry set, no borrow -> u3>=u1 -> ok
    return                          ; // B/ ----> done

    sub       w1, w2, w2            ; // u1 - u2 -> u2

    mov       #True, w0             ; // Assume in range
    btss      SR, #C                ; // S/ carry set, no borrow -> u2<=u1 -> ok
    mov       #False, w0            ; // Not in range
    return                          ; // B/ ----> done
  endCode


  // //////////////////////////////////////////////////////////////////////////

  code calc_crc ( nValue nOldCrc nPolyNomial -- nNewCrc )

    PopReg    w3                    ; // nPolynomial into w3
    PopReg    w4                    ; // nOldCrc into w4
    mov       w0, w1                ; // nValue into w1
    mov       #8, w2                ; // 8 bits to process, w2 is counter

1:
    ; // Process next bit

    xor       w4, w1, w0            ; // Xor nOldCrc with data into w0
    lsr       w4, w4                ; // Shift CRC right
    lsr       w1, w1                ; // Drop a bit of data

    btsc      w0, #0                ; // S/ bit 0 clear, no xor needed
    xor       w3, w4, w4            ; // Bit set, xor CRC with polynomial

    ; // Check if more bits to do

    dec       w2, w2                ; // One less bit to do
    bra       nz, 1b                ; // B/ more bits to do

    ; // Done

    mov       w4, w0                ; // Result to TOS
    return                          ; // B/ ----> done

  endCode
    #//
    #// Generalized CRC calculation of CRC's from 1 to 16 bits.
    #// The polynomial has to be specified in it's reversed form.


  // //////////////////////////////////////////////////////////////////////////

  : crc8 ( nValue nCrc -- nCrc )

    #//   Dallas crc-8, reversed polynomial $8c 1-wire bus
    #//   x^8 + x^5 + x^4 + x^0
    #//
    #// reentrant

    $8c calc_crc
  ;

  // //////////////////////////////////////////////////////////////////////////

  : crc16 ( nValue nCrc -- nCrc )

    #//  CRC-16 (IBM), reversed polynomial 0xa001
    #//  x^16 + x^15 + x^2 + x^0
    #//
    #// reentrant

    $a001 calc_crc
  ;

  // //////////////////////////////////////////////////////////////////////////

  : calcCrc8 ( nValue -- )

    #// Incorporate n into 8 bit CRC
    #//
    #//   Dallas crc-8, polynomial $8c
    #//   x^8 + x^5 + x^4 + x^0
    #//
    #// non - reentrant

    CRC@ crc8 CRC!
  ;


  // //////////////////////////////////////////////////////////////////////////

  : calcCrc16 ( nValue -- )

    #// Incorporate n into 16 bit CRC
    #//
    #//  CRC-16, polynomial 0xa001
    #//  x^16 + x^15 + x^2 + x^0
    #//
    #// non - reentrant

    CRC@  crc16 CRC!
  ;


  // //////////////////////////////////////////////////////////////////////////

  : dupCalcCrc8 ( u -- u )

    #// Incorporate u into 8 bit CRC
    #// non - reentrant

    dup calcCrc8
  ;


  // //////////////////////////////////////////////////////////////////////////

  : dupCalcCrc16 ( n -- n )

    #// Incorporate n into 16 bit CRC
    #// non - reentrant

    dup calcCrc16
  ;


  // //////////////////////////////////////////////////////////////////////////

  : blockCrc8Nc ( nCrc nCount nRamAddr -- nCrc )

    #// Calculate 8 bit CRC over memory block, not clearing CRC
    #// non - reentrant

    ramPtr!
    0 ?do
      ramc@++
      swap crc8
    loop
  ;


  // //////////////////////////////////////////////////////////////////////////

  : blockCrc16Nc ( nCrc nCount nRamAddr -- nCrc )

    #// Calculate 16 bit CRC over memory block, not clearing CRC
    #// non - reentrant

    ramPtr!
    0 ?do
      ramc@++
      swap crc16
    loop
  ;


  // //////////////////////////////////////////////////////////////////////////

  : blockCrc8 ( nCount nRamAddr -- nCrc )

    #// Calculate 8 bit CRC over memory block, clearing CRC
    #// non - reentrant

    0                               // Clear crc to be seed value
    blockCrc8Nc
  ;


  // //////////////////////////////////////////////////////////////////////////

  : blockCrc16 ( nCount nRamAddr -- nCRC )

    #// Calculate 16 bit CRC over memory block, clearing CRC
    #// non - reentrant

    0                               // Clear crc to be seed value
    blockCrc16Nc
  ;


  // //////////////////////////////////////////////////////////////////////////

  macro indexToMask ( ) ( u -- u ) // Make mask from index, valid indices : $00 .. $0f
                                    ; // indexToMask ( u -- u )
    mov       #1, w1                ; // Initial mask is 1
    sl        w1, w0, w0            ; // Shift it left over index positions
  endMacro


  // //////////////////////////////////////////////////////////////////////////

  code dIndexToMask ( u -- ud ) // Make mask from index, valid indices : $00 .. $1fh

    mov       #1, w1                ; // Initial mask is 1
    cp        w0, #0x0f             ; // Check w0 > $f
    bra       gt, 1f                ; // B/ w0 > 15

    ; // w0 <= 15

    sl        w1, w0, w0            ; // Shift it left over index positions
    clr       w1                    ; // Clear high byte
    bra       2f                    ; // B/ done
1:
    ; // w0 > 15

    sub       #0x10, w0             ; // Reduce problem modulo 16
    sl        w1, w0, w1            ; // Shift it left over index positions
    clr       w0                    ; // Clear low byte
2:
    PushReg   w1                    ; // Push high result over low
    return
  endCode


  // //////////////////////////////////////////////////////////////////////////

  code maskCheck ( nValue nMask -- f ) // Return true if (bValue and bMask) <> 0 : MaskCheck ( n n - f ) and 0<> ;

    and       w0, [w9--], w0        ; // (and) ( u u -- u )
    bra       z, 1f                 ; // B/ result is zero, is false, is what we want
    setm      w0                    ; // result not zero, set true to TOS
1:
    return
  endCode


  // //////////////////////////////////////////////////////////////////////////

  : dmaskCheck ( dValue dMask -- f )

    #// Return true if (bValue and bMask) <> 0

    dand d0= not
  ;


  // //////////////////////////////////////////////////////////////////////////

  todo' test bit operators'

  code b@ ( nBit nRamAddr -- f ) // Bit get

    PopReg    w1                    ; // Address into w1

    mov       [w1], w1              ; // nBit into w1 (index)
    mov       #1, w2                ; // Initial mask is 1
    sl        w2, w0, w2            ; // Shift it left over index positions
    and       w2, w1, w0            ; // And mask with value
    bra       z, 1f                 ; // B/ result is zero, is false, is what we want

    setm      w0                    ; // result not zero, set true to TOS
1:
    return                          ; // B/ ----> done

  endCode


  // //////////////////////////////////////////////////////////////////////////

  code b! ( f nBit nRamAddr -- ) // bit set

    PopReg    w2                    ; // Address into w2
    PopReg    w1                    ; // nBit into w1 (index)
                                    ; // Keep f in w0
    mov       #1, w3                ; // Initial mask is 1
    sl        w3, w1, w3            ; // Shift it left over index positions
    cp0       w0                    ; // Compare w0 against false
    bra       z, 1f                 ; // B/ w0 = false

    ; // w0 was true, or in bit

    ior       w3, [w2], [w2]
    bra       2f                    ; // B/ done
1:
    ; // w0 was false, nand in complement of bit

    com       w3, w3
    and       w3, [w2], [w2]
2:
    Drop                            ; // Drop f
    return                          ; // B/ ----> done

  endCode


  // //////////////////////////////////////////////////////////////////////////

  code b~ ( nBit nRamAddr -- ) // bit toggle

    PopReg    w1                    ; // Address into w1

    mov       #1, w2                ; // Initial mask is 1
    sl        w2, w0, w2            ; // Shift it left over index positions
    xor       w2, [w1], [w1]        ; // Xor at address

    Drop                            ; // Drop nBit
    return                          ; // B/ ----> done
  endCode


  // //////////////////////////////////////////////////////////////////////////

  macro reg@ ( aReg ) ( -- n )
                                    ; // reg@
    pushreg   $aReg$
  endMacro



  // //////////////////////////////////////////////////////////////////////////

  macro reg! ( aReg ) ( n -- )
                                    ; // reg!
    popreg    $aReg$
  endMacro



  // //////////////////////////////////////////////////////////////////////////

  macro reg@drop ( aReg ) ( -- )
                                    ; // regdrop - read register and dump contents
    mov       $aReg$, w1
  endMacro



  // //////////////////////////////////////////////////////////////////////////

  macro bitClear ( aReg aBit ) ( -- )
                                    ; // bitClear
    bclr      $aReg$, #$aBit$
  endMacro


  // //////////////////////////////////////////////////////////////////////////

  macro bitSet ( aReg aBit ) ( -- )
                                    ; // bitSet
    bset      $aReg$, #$aBit$
  endMacro


  // //////////////////////////////////////////////////////////////////////////

  macro bitToggle ( aReg aBit ) ( -- )
                                    ; // bitToggle
    btg       $aReg$, #$aBit$
  endMacro


  // //////////////////////////////////////////////////////////////////////////

  code d<< ( u n -- u ) // double left shift ( 0 <= n <= 31 )

    PopReg    w2                    ; // n into w2
    cp        w2, #31               ; // Check n to be in range
    bra       gt, 2f                ; // B/ w2 > 31
1:
    sl        [w9], [w9]            ; // shift low word, 0 in, bit 15 to carry
    rlc       w0, w0                ; // rotate high word through carry

    dec       w2, w2
    bra       nz, 1b                ; // B/ not done
    bra       3f                    ; // B/ done
2:
    clr       [w9]                  ; // w2 > 31, clear result
    clr       w0
3:
    return                          ; // B/ ----> done
  endCode


  // //////////////////////////////////////////////////////////////////////////

  code d>> ( u n -- u ) // double right shift ( 0 <= n <= 31 )

    PopReg    w2                    ; // n into w2
    cp        w2, #31               ; // Check n to be in range
    bra       gt, 2f                ; // B/ w2 > 31
1:
    lsr       w0, w0                ; // Shift high word, shifting in a zero
    rrc       [w9], [w9]            ; // rotate low word through carry

    dec       w2, w2
    bra       nz, 1b                ; // B/ not done
    bra       3f                    ; // B/ done
2:
    clr       [w9]                  ; // w2 > 31, clear result
    clr       w0
3:
    return                          ; // B/ ----> done
  endCode


  // //////////////////////////////////////////////////////////////////////////

  macro >< ( ) ( n1 -- n2 ) // Swap high and low byte of word
    swap      w0                    ; // >< ( n1 -- n2 ), swap high and low bytes
  endMacro


  // //////////////////////////////////////////////////////////////////////////

  macro c>< ( ) ( n1 -- n2 ) // Swap high and low nibbles of low byte
    swap.b    w0                    ; // >< ( n1 -- n2 ), swap high and low nibbles in low byte
  endMacro


  // //////////////////////////////////////////////////////////////////////////
  //
  // from hlpMPLABC30.chm :
  //
  //   The 16-Bit Device Compiler has the following managed resources: W0-W15,
  //   RCOUNT, SR.
  //
  // And from 70157 - programmers reference manual.pdf
  //
  //   Table 2-2: Automatic Shadow Register Usage
  //
  //     Location                   | DO    | POP.S/PUSH.S
  //     ---------------------------+-------+-------------
  //     DCOUNT                     | Yes   |  
  //     DOSTART                    | Yes   |  
  //     DOEND                      | Yes   |  
  //     STATUS Register            |       |
  //       DC, N, OV, Z and C bits  |      |  Yes
  //     W0-W3                      |      |  Yes
  //
  //   push.s =
  //
  //     push w0 .. w3 and sr into shadow set
  //
  //   push reg =
  //
  //     mov    reg, [w15++]
  //
  //   pop  reg =
  //
  //     mov    [--w15], reg


  macro interruptProlog ( ) ( -- )
                                    ; // +----------------+---------------------+
                                    ; // | value  / 4th's | offset after prolog |
                                    ; // +----------------+---------------------+
                                    ; // | Ret. addr.     | FP - 36  (dword)    |
    push      RCOUNT                ; // | RCOUNT         | FP - 32             |
    push.d    w0                    ; // | w0     / TOS   | FP - 30             |
                                    ; // | w1             | FP - 28             |
    push.d    w2                    ; // | w2             | FP - 26             |
                                    ; // | w3             | FP - 24             |
    push.d    w4                    ; // | w4             | FP - 22             |
                                    ; // | w5             | FP - 20             |
    push.d    w6                    ; // | w6             | FP - 18             |
                                    ; // | w7             | FP - 16             |
    push.d    w8                    ; // | w8             | FP - 14             |
                                    ; // | w9     / DSP   | FP - 12             |
    push.d    w10                   ; // | w10    / LSP   | FP - 10             |
                                    ; // | w11            | FP -  8             |
    push.d    w12                   ; // | w12            | FP -  6             |
                                    ; // | w13    / EFP   | FP -  4             |
    lnk       #0                    ; // | w14            | FP -  2             |
                                    ; // | undefined      | <--- FP, SP         |
                                    ; // +----------------+---------------------+

    rcall     <@" forthInterruptSetup">
  endMacro


  // //////////////////////////////////////////////////////////////////////////

  macro interruptEpilog ( ) ( -- )
    ulnk
    pop.d     w12
    pop.d     w10
    pop.d     w8
    pop.d     w6
    pop.d     w4
    pop.d     w2
    pop.d     w0
    pop       RCOUNT
  endMacro


  // //////////////////////////////////////////////////////////////////////////

  : random ( nSeed -- nResult )

    #// Simple linear congruential random generator, seeded.
    #//
    #// This one should be kept for compatibilty reasons with the PIC18
    #// versions, it is used there by the boot loader.

    $ff and                         // Mask the seed modulo 256
    >r                              // Keep it for later
    1103515245.                     // Multiplier
    r> u>d                          // Extend seed to double
    d* drop                         // Multiply and mask mod 65536
    12345 +                         // Add Addend
    $ff and                         // Mask result modulo 256
  ;


  // //////////////////////////////////////////////////////////////////////////

public

  export  ramPtr@           (                       -- naddr    ) // Read ramPtr
  export  ramPtr!           ( nRamAddr              --          ) // Setup ramPtr
  export  ramPtr++          (                       --          ) // Increment ramPtr
  export  ramPtr--          (                       --          ) // Decrement ramPtr
  export  ramc@             (                       -- n        ) // Get byte value indirect through ramPtr
  export  ramc!             ( n                     --          ) // Set byte value indirect through ramPtr
  export  ramc@++           (                       -- n        ) //
  export  ramc!++           ( n                     --          ) //
  export  ram++c@           (                       -- n        ) //
  export  ram++c!           ( n                     --          ) //
  export  ramc@--           (                       -- n        ) //
  export  ramc!--           ( n                     --          ) //
  export  ram--c@           (                       -- n        ) //
  export  ram--c!           ( n                     --          ) //

  export  romPtr!           ( dRomAddr              --          )
  export  romPtr++          (                       --          )
  export  romPtr--          (                       --          )
  export  romc@             (                       -- n        )
  export  romc!             ( n                     --          )
  export  romc@++           (                       -- n        )
  export  romc!++           ( n                     --          )
  export  rom++c@           (                       -- n        )
  export  rom++c!           ( n                     --          )
  export  romc@--           (                       -- n        )
  export  romc!--           ( n                     --          )
  export  rom--c@           (                       -- n        )
  export  rom--c!           ( n                     --          )

  export  inRange           ( u1 u2 u3              -- f        ) // True when u2 <= u1 <= u3

  export  crc@              (                       -- u        ) // Fetch CRC value as a word
  export  crc!              ( u                     --          ) // Store CRC value as a word

  export  crc8              ( uValue uCrc           -- uCrc     ) // Incorporate uValue into uCrc
  export  calcCrc8          ( u                     --          ) // Incorporate u into 8 bit CRC
  export  dupCalcCrc8       ( u                     -- u        ) // Incorporate u into 8 bit CRC
  export  blockCrc8Nc       ( uCrc uCount uRamAddr  -- uCrc     ) // Calculate 8 bit CRC over memory block, not clearing CRC
  export  blockCrc8         ( uCount uRamAddr       -- uCrc     ) // Calculate 8 bit CRC over memory block,     clearing CRC

  export  crc16             ( uValue uCrc           -- uCrc     ) // Incorporate uValue into uCrc
  export  calcCrc16         ( u                     --          ) // Incorporate u into 16 bit CRC
  export  blockCrc16Nc      ( uCrc uCount uRamAddr  --          ) // Calculate 16 bit CRC over memory block, not clearing CRC
  export  blockCrc16        ( uCount uRamAddr       -- wCRC     ) // Calculate 16 bit CRC over memory block,     clearing CRC

  export  dindexToMask      ( u                     -- ud       ) // Make mask from index, valid indices : 0x00 .. 0x1f
  export  indexToMask       ( u                     -- u        ) // Make mask from index, valid indices : 0x00 .. 0x0f
  export  dmaskCheck        ( uValue uMask          -- f        ) // Return true if (dValue and dMask) <> 0
  export  maskCheck         ( uValue uMask          -- f        ) // Return true if (bValue and bMask) <> 0

  export  b@                ( nBit nRamAddr         -- f        ) // bit get
  export  b!                ( f nBit nRamAddr       --          ) // bit set
  export  b~                ( nBit nRamAddr         --          ) // bit toggle
  export  reg@              (                       -- n        ) // reg fetch macro reg@ aReg
  export  reg!              ( n                     --          ) // reg store macro reg! aReg
  export  reg@drop          (                       --          ) // reg fetch macro reg@drop aReg (discard read results)
  export  bitClear          (                       --          ) // bit clear  macro bitClear  aReg aBit
  export  bitSet            (                       --          ) // bit set    macro bitSet    aReg aBit
  export  bitToggle         (                       --          ) // bit toggle macro bitToggle aReg aBit
  export  d<<               ( u n                   -- u        ) // double left  shift ( 0 <= n <= 31 )
  export  d>>               ( u n                   -- u        ) // double right shift ( 0 <= n <= 31 )
  export  ><                ( n1                    -- n2       ) // Swap high and low byte
  export  c><               ( n1                    -- n2       ) // Swap high and low nibbles of low byte

  export  interruptProlog   (                       --          ) // Standard interrupt prolog macro
  export  interruptEpilog   (                       --          ) // Standard interrupt epilog macro

  export  random            ( nSeed                 -- nResult  ) // Simple linear congruential random generator, seeded.

endModule

